import { Scene, PerspectiveCamera, WebGLRenderer, Raycaster, Group, TextureEncoding, Object3D, Clock, AnimationMixer, Box3Helper, Box3, Vector3, Intersection, Event } from 'three';
import mapboxgl from 'mapbox-gl';
import { CSS2DObject } from 'three/examples/jsm/renderers/CSS2DRenderer';
import CameraSync from './CameraSync';
import Utils from './Utils';
import Material from './Material';
import { EventHandler, ObjectHandler, AnimationHandler } from './handlers';
import EventDispatcher from './EventDispatcher';
export declare type ThreeboxOption = {
    /**
     * @default THREE.GammaEncoding
     * @type {TextureEncoding}
     */
    outputEncoding?: TextureEncoding;
    /**
     * @default 2
     * @type {number}
     */
    gammaFactor?: number;
    /**
     * Whether using map trigger paint
     *
     * @type {boolean}
     */
    passiveRendering?: boolean;
    /**
     * Whether using default lights on scene
     *
     * @type {boolean}
     */
    useDefaultLights?: boolean;
    /**
     * Whether using tooltip at hover
     *
     * @type {boolean}
     */
    useTooltip?: boolean;
    /**
     * Whether using hovering object and feature
     *
     * @type {boolean}
     */
    useHover?: boolean;
    /**
     * Whether using selecting object3D
     *
     * @type {boolean}
     */
    useSelectingObjects?: boolean;
    /**
     * Whether using selecting map 3d buildings feature
     *
     * @type {boolean}
     */
    useSelectingFeatures?: boolean;
    /**
     * Whether using rotating object3D
     *
     * @type {boolean}
     */
    useRotatingObjects?: boolean;
    /**
     * Whether using dragging object3D
     *
     * @type {boolean}
     */
    useDraggingObjects?: boolean;
};
export declare type ThreeboxEventType = 'select' | 'add' | 'remove' | 'moved' | 'moving' | 'scaled' | 'scaling' | 'rotated' | 'rotating';
export declare type LoadModelType = 'gltf' | 'obj' | 'fbx' | 'dae';
export interface LoadModelOption {
    /**
     * Model url
     *
     * @type {string}
     */
    url: string;
    /**
     * MTL url
     *
     * @type {string}
     */
    mtl?: string;
    /**
     * Load model type
     *
     * @type {LoadModelType}
     */
    type: LoadModelType;
    /**
     * @default meters
     * @type {('scene' | 'meters')}
     */
    units?: 'scene' | 'meters';
    /**
     * @default 1
     * @type {number}
     */
    scale?: number;
    /**
     * @default { x: 90, y: 0, z: 0 }
     * @type {(number | { x: number; y: number; z: number })}
     */
    rotation?: number | {
        x: number;
        y: number;
        z: number;
    };
    /**
     * @default 1
     * @type {number}
     */
    defaultAnimation?: number;
    /**
     * @default { x: 0, y: 0, z: 0 }
     * @type {{ x: number; y: number; z: number }}
     */
    adjustment?: {
        x: number;
        y: number;
        z: number;
    };
    /**
     * @default false
     * @type {boolean}
     */
    normalize?: boolean;
}
export declare type ThreeboxObjectEventType = 'changed.wireframe' | 'changed.selected' | 'changed.isPlyaing' | 'object.dragged' | 'object.mouseover' | 'object.mouseout';
export interface ThreeboxObject extends Omit<Object3D, 'add' | 'remove'> {
    isPlaying?: boolean;
    animationQueue?: any[];
    animationMethod?: any;
    clock?: Clock;
    hasDefaultAnimation?: boolean;
    defaultAction?: any;
    actions?: any[];
    mixer?: AnimationMixer;
    animations?: any[];
    /**
     * @example [Longitude, Latitude, Altitude?]
     * @type {number[]}
     */
    coordinates?: number[];
    modelHeight?: number;
    modelSize?: Vector3;
    boundingBox?: Box3Helper;
    boundingBoxShadow?: Box3Helper;
    label?: CSS2DObject & {
        alwaysVisible?: boolean;
        dispose?: () => void;
    };
    tooltip?: any;
    visibility?: boolean;
    model?: Group;
    wireframe?: boolean;
    selected?: boolean;
    hover?: boolean;
    set?: (option: any) => void;
    playDefault?: (option: any) => void;
    playAnimation?: (option: any) => void;
    pauseAllActions?: () => void;
    resumeAllActions?: () => void;
    stop?: (index?: number) => void;
    deactivateAllActions?: () => void;
    activateAllActions?: () => void;
    idle?: () => void;
    followPath?: (option: any, callback?: any) => void;
    setObject?: (option: any) => void;
    /**
     * @param lnglat [Longitude, Latitude, Altitude?]
     */
    setCoords?: (lnglat: number[]) => void;
    setBoundingBoxShadowFloor?: () => void;
    /**
     * @param lnglat [Longitude, Latitude, Altitude?]
     */
    setTranslate?: (lnglat: number[]) => void;
    setRotation?: (xyz: number | {
        x?: number;
        y?: number;
        z: number;
    }) => void;
    setRotationAxis?: (xyz: number | {
        x?: number;
        y?: number;
        z: number;
    }) => void;
    /**
     * @param lnglat [Longitude, Latitude, Altitude?]
     * @param xyz
     * @param inverse
     */
    calcAdjustedPosition?: (lnglat: number[], xyz: {
        x: number;
        y: number;
        z: number;
    }, inverse?: boolean) => void;
    drawBoundingBox?: () => Group;
    addLabel?: (element: HTMLElement, visible?: boolean) => void;
    drawLabelHTML?: (element: HTMLElement, visible?: boolean) => CSS2DObject;
    addTooltip?: (text: string, mapboxStyle?: boolean, center?: {
        x: number;
        y: number;
        z: number;
    }) => void;
    box3?: () => Box3;
    modelBox?: () => Box3;
    getSize?: () => Vector3;
    add?: () => void;
    remove?: () => void;
    duplicate?: () => this;
    dispose?: () => void;
    /**
     * Adds a listener to an event type.
     * @param {ThreeboxObjectEventType} type The type of event to listen to.
     * @param {(event: Event) => void} listener The function that gets called when the event is fired.
     */
    addEventListener(type: ThreeboxObjectEventType, listener: (event: Event) => void): void;
    /**
     * Checks if listener is added to an event type.
     * @param {ThreeboxObjectEventType} type The type of event to listen to.
     * @param {(event: Event) => void} listener The function that gets called when the event is fired.
     */
    hasEventListener(type: ThreeboxObjectEventType, listener: (event: Event) => void): boolean;
    /**
     * Removes a listener from an event type.
     * @author salgum1114
     * @param {ThreeboxObjectEventType} type The type of the listener that gets removed.
     * @param {(event: Event) => void} listener The listener function that gets removed.
     */
    removeEventListener(type: ThreeboxObjectEventType, listener: (event: Event) => void): void;
    /**
     * Fire an event type.
     * @param type The type of event that gets fired.
     */
    dispatchEvent(event: {
        type: ThreeboxObjectEventType;
        [attachment: string]: any;
    }): void;
}
export declare const ThreeboxDefaultOption: ThreeboxOption;
export declare const LoadModelDefaultOption: LoadModelOption;
declare class Threebox extends EventDispatcher {
    version: string;
    map: mapboxgl.Map;
    gl: WebGLRenderingContext;
    scene: Scene;
    camera: PerspectiveCamera;
    renderer: WebGLRenderer;
    raycaster: Raycaster;
    world: Group;
    objectsCache: any[];
    cameraSync: CameraSync;
    labelRenderer: any;
    option: ThreeboxOption;
    useTooltip?: boolean;
    useDefaultLights?: boolean;
    useHover?: boolean;
    useSelectingObjects?: boolean;
    useSelectingFeatures?: boolean;
    useRotatingObjects?: boolean;
    useDraggingObjects?: boolean;
    utils: typeof Utils;
    material: typeof Material;
    events: EventHandler;
    objects: ObjectHandler;
    animations: AnimationHandler;
    /**
     * Creates an instance of Threebox.
     * @author salgum1114
     * @param {mapboxgl.Map} map
     * @param {WebGLRenderingContext} gl
     * @param {ThreeboxOption} [option]
     */
    constructor(map: mapboxgl.Map, gl: WebGLRenderingContext, option?: ThreeboxOption);
    /**
     * Set up event option
     *
     * @author salgum1114
     * @param {ThreeboxOption} option
     */
    setOption(option: ThreeboxOption): void;
    /**
     *
     *
     * @author salgum1114
     */
    defaultLights(): void;
    /**
     *
     *
     * @author salgum1114
     * @returns
     */
    memory(): {
        geometries: number;
        textures: number;
    };
    /**
     *
     *
     * @author salgum1114
     * @returns
     */
    programs(): number;
    /**
     *
     *
     * @author salgum1114
     * @param {[number, number, number]} coords
     * @returns
     */
    projectToWorld(coords: [number, number, number]): Vector3;
    /**
     *
     *
     * @author salgum1114
     * @param {{ x: number; y: number; z: number }} v3
     * @returns
     */
    unprojectFromWorld(v3: {
        x: number;
        y: number;
        z: number;
    }): number[];
    /**
     *
     *
     * @author salgum1114
     * @param {number} lat
     * @returns
     */
    projectedUnitsPerMeter(lat: number): number;
    /**
     * get the center point of a feature
     *
     * @author salgum1114
     * @param {GeoJSON.Feature<any>} feature
     * @param {ThreeboxObject} obj
     * @param {number} level
     * @returns
     */
    getFeatureCenter(feature: GeoJSON.Feature<any>, obj?: ThreeboxObject, level?: number): number[];
    /**
     *
     *
     * @author salgum1114
     * @param {GeoJSON.Feature<any>} feature
     * @param {ThreeboxObject} [obj]
     * @param {number} [level]
     * @returns
     */
    getObjectHeightOnFloor(feature: GeoJSON.Feature<any>, obj?: ThreeboxObject, level?: number): number;
    /**
     *
     *
     * @author salgum1114
     * @param {mapboxgl.Point} point
     * @returns
     */
    queryRenderedFeatures(point: mapboxgl.Point): Intersection[];
    /**
     * find 3D object of a mesh. this method is needed to know the object of a raycasted mesh
     *
     * @author salgum1114
     * @param {Intersection} mesh
     * @returns
     */
    findParent3DObject(mesh: Intersection): ThreeboxObject;
    /**
     * method to replicate behaviour of map.setLayoutProperty when Threebox are affected
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {string} name
     * @param {*} value
     * @returns
     */
    setLayoutProperty(layerId: string, name: string, value: any): void;
    /**
     * Custom Layers doesn't work on minzoom and maxzoom attributes, and if the layer is including labels they don't hide either on minzoom
     *
     * @author salgum1114
     * @param {string} layer3d
     * @param {number} minZoomLayer
     * @param {number} maxZoomLayer
     */
    setLayerZoomRange(layer3d: string, minZoomLayer: number, maxZoomLayer: number): void;
    /**
     * method to set the height of all the objects in a level. this only works if the objects have a geojson feature
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {number} level
     * @returns
     */
    setLayerHeigthProperty(layerId: string, level: number): void;
    /**
     * method to toggle Layer visibility
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {boolean} visible
     */
    toggleLayer(layerId: string, visible: boolean): void;
    /**
     * method set the CSS2DObjects zoom range and hide them at the same time the layer is
     *
     * @deprecated
     * @author salgum1114
     * @param {number} minZoom
     * @param {number} maxZoom
     */
    setLabelZoomRange(minZoom: number, maxZoom: number): void;
    /**
     * Update scene
     *
     * @author salgum1114
     */
    update(): void;
    add(obj: any): void;
    remove(obj: any): void;
    /**
     * method to fully dispose the resources, watch out is you call this without navigating to other page
     *
     * @author salgum1114
     * @returns
     */
    dispose(): Promise<boolean>;
    loadModel(option: LoadModelOption): Promise<ThreeboxObject>;
}
export default Threebox;
