"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LoadModelDefaultOption = exports.ThreeboxDefaultOption = void 0;
const three_1 = require("three");
const OBJLoader_1 = require("three/examples/jsm/loaders/OBJLoader");
const GLTFLoader_1 = require("three/examples/jsm/loaders/GLTFLoader");
const FBXLoader_1 = require("three/examples/jsm/loaders/FBXLoader");
const ColladaLoader_1 = require("three/examples/jsm/loaders/ColladaLoader");
const MTLLoader_1 = require("three/examples/jsm/loaders/MTLLoader");
const CameraSync_1 = __importDefault(require("./CameraSync"));
const Constant_1 = __importDefault(require("./Constant"));
const Utils_1 = __importDefault(require("./Utils"));
const Material_1 = __importDefault(require("./Material"));
const handlers_1 = require("./handlers");
const EventDispatcher_1 = __importDefault(require("./EventDispatcher"));
exports.ThreeboxDefaultOption = {
    outputEncoding: three_1.GammaEncoding,
    gammaFactor: 2,
    passiveRendering: true,
    useDefaultLights: true,
    useTooltip: false,
    useHover: false,
    useSelectingObjects: false,
    useSelectingFeatures: false,
    useRotatingObjects: false,
    useDraggingObjects: false,
};
exports.LoadModelDefaultOption = {
    url: null,
    type: null,
    units: 'meters',
    scale: 1,
    rotation: { x: 90, y: 0, z: 0 },
    defaultAnimation: 0,
    adjustment: { x: 0, y: 0, z: 0 },
    normalize: false,
};
class Threebox extends EventDispatcher_1.default {
    /**
     * Creates an instance of Threebox.
     * @author salgum1114
     * @param {mapboxgl.Map} map
     * @param {WebGLRenderingContext} gl
     * @param {ThreeboxOption} [option]
     */
    constructor(map, gl, option) {
        super();
        this.version = VERSION;
        // Threebox utils
        this.utils = Utils_1.default;
        this.map = map;
        this.gl = gl;
        // apply start option
        this.option = this.utils.validate(option || {}, exports.ThreeboxDefaultOption);
        this.material = Material_1.default;
        this.objects = new handlers_1.ObjectHandler(this, map);
        this.events = new handlers_1.EventHandler(this, map);
        this.animations = new handlers_1.AnimationHandler(this, map);
        // Set up WebGLRenderer
        this.renderer = new three_1.WebGLRenderer({
            canvas: map.getCanvas(),
            antialias: true,
            context: gl,
        });
        this.renderer.setPixelRatio(globalThis.devicePixelRatio);
        this.renderer.setSize(map.getCanvas().clientWidth, map.getCanvas().clientHeight);
        this.renderer.autoClear = false;
        this.renderer.outputEncoding = option?.outputEncoding || three_1.GammaEncoding;
        this.renderer.gammaFactor = option?.gammaFactor || 2;
        // set labelRendered
        // this.labelRenderer = new LabelRenderer(this.map);
        // create Scene
        this.scene = new three_1.Scene();
        // create Perspective camera
        this.camera = new three_1.PerspectiveCamera(Constant_1.default.FOV_DEGREES, map.getCanvas().clientWidth / map.getCanvas().clientHeight, 1, 1e21);
        this.camera.layers.enable(0);
        this.camera.layers.enable(1);
        // The CameraSync object will keep the Mapbox and THREE.js camera movements in sync.
        // It requires a world group to scale as we zoom in. Rotation is handled in the camera's
        // projection matrix itself (as is field of view and near/far clipping)
        // It automatically registers to listen for move events on the map so we don't need to do that here
        this.world = new three_1.Group();
        this.scene.add(this.world);
        this.objectsCache = [];
        this.cameraSync = new CameraSync_1.default(this.map, this.camera, this.world);
        // raycaster for mouse events
        this.raycaster = new three_1.Raycaster();
        this.raycaster.layers.set(0);
        // set up event option
        this.setOption(this.option);
        // create default lights
        this.defaultLights();
    }
    /**
     * Set up event option
     *
     * @author salgum1114
     * @param {ThreeboxOption} option
     */
    setOption(option) {
        // TODO...
        console.debug(this.utils.validate(option || {}, this));
        this.useDefaultLights = option.useDefaultLights;
        this.useTooltip = option.useTooltip;
        this.useHover = option.useHover;
        this.useSelectingFeatures = option.useSelectingFeatures;
        this.useSelectingObjects = option.useSelectingObjects;
        this.useDraggingObjects = option.useDraggingObjects;
    }
    /**
     *
     *
     * @author salgum1114
     */
    defaultLights() {
        if (this.useDefaultLights) {
            const ambientLight = new three_1.AmbientLight(new three_1.Color('hsl(0, 0%, 100%)'), 0.75);
            this.scene.add(ambientLight);
            const directionalLightBack = new three_1.DirectionalLight(new three_1.Color('hsl(0, 0%, 100%)'), 0.25);
            directionalLightBack.position.set(10, 100, 100);
            this.scene.add(directionalLightBack);
            const directionalLightFront = new three_1.DirectionalLight(new three_1.Color('hsl(0, 0%, 100%)'), 0.25);
            directionalLightFront.position.set(-10, -100, 100);
            this.scene.add(directionalLightFront);
        }
    }
    /**
     *
     *
     * @author salgum1114
     * @returns
     */
    memory() {
        return this.renderer.info.memory;
    }
    /**
     *
     *
     * @author salgum1114
     * @returns
     */
    programs() {
        return this.renderer.info.programs.length;
    }
    /**
     *
     *
     * @author salgum1114
     * @param {[number, number, number]} coords
     * @returns
     */
    projectToWorld(coords) {
        return this.utils.projectToWorld(coords);
    }
    /**
     *
     *
     * @author salgum1114
     * @param {{ x: number; y: number; z: number }} v3
     * @returns
     */
    unprojectFromWorld(v3) {
        return this.utils.unprojectFromWorld(v3);
    }
    /**
     *
     *
     * @author salgum1114
     * @param {number} lat
     * @returns
     */
    projectedUnitsPerMeter(lat) {
        return this.utils.projectedUnitsPerMeter(lat);
    }
    /**
     * get the center point of a feature
     *
     * @author salgum1114
     * @param {GeoJSON.Feature<any>} feature
     * @param {ThreeboxObject} obj
     * @param {number} level
     * @returns
     */
    getFeatureCenter(feature, obj, level) {
        return this.utils.getFeatureCenter(feature, obj, level);
    }
    /**
     *
     *
     * @author salgum1114
     * @param {GeoJSON.Feature<any>} feature
     * @param {ThreeboxObject} [obj]
     * @param {number} [level]
     * @returns
     */
    getObjectHeightOnFloor(feature, obj, level) {
        return this.utils.getObjectHeightOnFloor(feature, obj, level);
    }
    /**
     *
     *
     * @author salgum1114
     * @param {mapboxgl.Point} point
     * @returns
     */
    queryRenderedFeatures(point) {
        const map = this.map;
        const mouse = new three_1.Vector2();
        // // scale mouse pixel position to a percentage of the screen's width and height
        mouse.x = (point.x / map.transform.width) * 2 - 1;
        mouse.y = 1 - (point.y / map.transform.height) * 2;
        this.raycaster.setFromCamera(mouse, this.camera);
        // calculate objects intersecting the picking ray
        const intersects = this.raycaster.intersectObjects(this.world.children, true);
        return intersects;
    }
    /**
     * find 3D object of a mesh. this method is needed to know the object of a raycasted mesh
     *
     * @author salgum1114
     * @param {Intersection} mesh
     * @returns
     */
    findParent3DObject(mesh) {
        // find the Parent Object3D of the mesh captured by Raytracer
        let result;
        mesh.object.traverseAncestors((m) => {
            if (m.parent) {
                if (m.parent.type === 'Group' && m.userData.url) {
                    result = m;
                }
            }
        });
        return result;
    }
    /**
     * method to replicate behaviour of map.setLayoutProperty when Threebox are affected
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {string} name
     * @param {*} value
     * @returns
     */
    setLayoutProperty(layerId, name, value) {
        // first set layout property at the map
        this.map.setLayoutProperty(layerId, name, value);
        if (value !== null && value !== undefined) {
            if (name === 'visibility') {
                this.world.children.forEach((obj) => {
                    if (obj.userData.feature && obj.userData.feature.layer === layerId) {
                        obj.visibility = value;
                    }
                });
                return;
            }
        }
    }
    /**
     * Custom Layers doesn't work on minzoom and maxzoom attributes, and if the layer is including labels they don't hide either on minzoom
     *
     * @author salgum1114
     * @param {string} layer3d
     * @param {number} minZoomLayer
     * @param {number} maxZoomLayer
     */
    setLayerZoomRange(layer3d, minZoomLayer, maxZoomLayer) {
        if (this.map.getLayer(layer3d)) {
            this.map.setLayerZoomRange(layer3d, minZoomLayer, maxZoomLayer);
            this.setLabelZoomRange(minZoomLayer, maxZoomLayer);
        }
    }
    /**
     * method to set the height of all the objects in a level. this only works if the objects have a geojson feature
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {number} level
     * @returns
     */
    setLayerHeigthProperty(layerId, level) {
        const map = this.map;
        const layer = map.getLayer(layerId);
        if (!layer)
            return;
        if (layer.type === 'fill-extrusion') {
            const data = map.getStyle().sources[layer.source].data;
            const features = data.features;
            features.forEach(f => {
                f.properties.level = level;
            });
            // we change the level on the source
            map.getSource(layer.source).setData(data);
        }
        else if (layer.type === 'custom') {
            this.world.children.forEach((obj) => {
                const feature = obj.userData.feature;
                if (feature && feature.layer === layerId) {
                    // TODO...
                    // this could be a multidimensional array
                    const location = this.utils.getFeatureCenter(feature, obj, level);
                    obj.setCoords(location);
                }
            });
        }
    }
    /**
     * method to toggle Layer visibility
     *
     * @author salgum1114
     * @param {string} layerId
     * @param {boolean} visible
     */
    toggleLayer(layerId, visible) {
        if (this.map.getLayer(layerId)) {
            this.setLayoutProperty(layerId, 'visibility', visible ? 'visible' : 'none');
        }
    }
    /**
     * method set the CSS2DObjects zoom range and hide them at the same time the layer is
     *
     * @deprecated
     * @author salgum1114
     * @param {number} minZoom
     * @param {number} maxZoom
     */
    setLabelZoomRange(minZoom, maxZoom) {
        console.warn('[DEPRECATED] Not used', minZoom, maxZoom);
        // this.labelRenderer.setZoomRange(minZoom, maxZoom);
    }
    /**
     * Update scene
     *
     * @author salgum1114
     */
    update() {
        if (this.map.repaint) {
            this.map.repaint = false;
        }
        const timestamp = Date.now();
        // Update any animations
        this.animations.update(timestamp);
        this.renderer.state.reset();
        // Render the scene and repaint the map
        this.renderer.render(this.scene, this.camera);
        // Render any label
        // this.labelRenderer.render(this.scene, this.camera);
        if (this.option.passiveRendering === false) {
            this.map.triggerRepaint();
        }
    }
    add(obj) {
        // remove the tooltip if not enabled
        if (!this.useTooltip && obj.tooltip) {
            obj.tooltip.visibility = false;
        }
        this.world.add(obj);
    }
    remove(obj) {
        // remove also the label if exists dispatching the event removed to fire CSS2DRenderer "removed" listener
        if (obj.label) {
            obj.label.remove();
        }
        if (obj.tooltip) {
            obj.tooltip.remove();
        }
        this.world.remove(obj);
    }
    /**
     * method to fully dispose the resources, watch out is you call this without navigating to other page
     *
     * @author salgum1114
     * @returns
     */
    dispose() {
        console.log(this.memory());
        return new Promise(resolve => {
            this.world.traverse((obj) => {
                if (obj.geometry) {
                    obj.geometry.dispose();
                }
                if (obj.material) {
                    if (obj.material instanceof three_1.Mesh) {
                        obj.material.materials.forEach((m) => {
                            m.dispose();
                            if (m.map) {
                                m.map.dispose();
                            }
                        });
                    }
                    else {
                        obj.material.dispose();
                    }
                    const m = obj.material;
                    const md = m.map ||
                        m.alphaMap ||
                        m.aoMap ||
                        m.bumpMap ||
                        m.displacementMap ||
                        m.emissiveMap ||
                        m.envMap ||
                        m.lightMap ||
                        m.metalnessMap ||
                        m.normalMap ||
                        m.roughnessMap;
                    if (md) {
                        if (m.map) {
                            m.map.dispose();
                        }
                        if (m.alphaMap) {
                            m.alphaMap.dispose();
                        }
                        if (m.aoMap) {
                            m.aoMap.dispose();
                        }
                        if (m.bumpMap) {
                            m.bumpMap.dispose();
                        }
                        if (m.displacementMap) {
                            m.displacementMap.dispose();
                        }
                        if (m.emissiveMap) {
                            m.emissiveMap.dispose();
                        }
                        if (m.envMap) {
                            m.envMap.dispose();
                        }
                        if (m.lightMap) {
                            m.lightMap.dispose();
                        }
                        if (m.metalnessMap) {
                            m.metalnessMap.dispose();
                        }
                        if (m.normalMap) {
                            m.normalMap.dispose();
                        }
                        if (m.roughnessMap) {
                            m.roughnessMap.dispose();
                        }
                    }
                }
                if (obj.dispose) {
                    obj.dispose();
                }
            });
            // this.map.remove();
            // this.map = {};
            this.scene.remove(this.world);
            this.scene.dispose();
            this.world.children = [];
            this.world = null;
            // this.labelRenderer.dispose();
            console.log(this.memory());
            this.renderer.dispose();
            this.events.dispose();
            resolve(true);
        });
    }
    async loadModel(option) {
        if (!option.type) {
            console.warn('[WARNING] Required type option', option);
            return Promise.reject();
        }
        const _option = this.utils.validate(option, exports.LoadModelDefaultOption);
        let loader;
        // support other models
        switch (_option.type) {
            case 'obj':
                // TODO: Support formats other than OBJ/MTL
                loader = new OBJLoader_1.OBJLoader();
                break;
            case 'gltf':
                // Support for GLTF
                loader = new GLTFLoader_1.GLTFLoader();
                break;
            case 'fbx':
                loader = new FBXLoader_1.FBXLoader();
                break;
            case 'dae':
                loader = new ColladaLoader_1.ColladaLoader();
                break;
        }
        if (_option.type === 'obj') {
            const mtlLoader = new MTLLoader_1.MTLLoader();
            const materials = (await mtlLoader.loadAsync(_option.mtl));
            if (materials && _option.type === 'obj') {
                materials.preload();
                const objLoader = loader;
                objLoader.setMaterials(materials);
            }
        }
        const obj = await loader.loadAsync(_option.url);
        let group = obj;
        // MTL/GLTF/FBX models have a different structure
        let animations;
        switch (_option.type) {
            case 'obj':
                group = obj.children[0];
                break;
            case 'gltf':
            case 'dae':
                animations = obj.animations;
                group = obj.scene;
                break;
            case 'fbx':
                animations = obj.animations;
                break;
        }
        // options.rotation was wrongly used
        const r = this.utils.rotation(_option.rotation, [0, 0, 0]);
        const s = this.utils.scale(_option.scale, [1, 1, 1]);
        group.rotation.set(r[0], r[1], r[2]);
        group.scale.set(s[0], s[1], s[2]);
        // normalize specular/metalness/shininess from meshes in FBX and GLB model as it would need 5 lights to illuminate them properly
        if (_option.normalize) {
            normalizeSpecular(group);
        }
        const projScaleGroup = new three_1.Group();
        projScaleGroup.add(group);
        const userScaleGroup = this.objects.makeGroup(projScaleGroup, _option);
        userScaleGroup.model = group;
        // assign the animations to the userScaleGroup before enrolling it in AnimationsManager through _addMethods
        userScaleGroup.animations = animations;
        this.objects.addMethods(userScaleGroup);
        // if the object options have an adjustment to center the 3D Object
        const center = _option.adjustment;
        if (center) {
            const size = userScaleGroup.getSize();
            group.position.set(size.x * center.x, size.y * center.y, size.z * center.z);
        }
        // after adding methods create the bounding box at userScaleGroup but add it to its children for positioning
        const boxGrid = userScaleGroup.drawBoundingBox();
        projScaleGroup.add(boxGrid);
        // we add by default a tooltip that can be override later or hide it with threebox `enableTooltips`
        userScaleGroup.addTooltip(userScaleGroup.uuid, true, center);
        // initialize the default animation to avoid issues with position
        userScaleGroup.idle();
        // some FBX/GLTF models have too much specular effects for mapbox
        function normalizeSpecular(model) {
            model.traverse((c) => {
                if (c.isMesh) {
                    let specularColor;
                    if (c.material.type === 'MeshStandardMaterial') {
                        if (c.material.metalness) {
                            c.material.metalness *= 0.1;
                        }
                        if (c.material.glossiness) {
                            c.material.glossiness *= 0.25;
                        }
                        specularColor = new three_1.Color(12, 12, 12);
                    }
                    else if (c.material.type === 'MeshPhongMaterial') {
                        c.material.shininess = 0.1;
                        specularColor = new three_1.Color(20, 20, 20);
                    }
                    if (c.material.specular && c.material.specular.isColor) {
                        c.material.specular = specularColor;
                    }
                }
            });
        }
        return userScaleGroup;
    }
}
exports.default = Threebox;
